from collections import abc as abc
from typing import Any, TypeVar

_R = TypeVar("_R")
_T = TypeVar("_T")
_R_Co = TypeVar("_R_Co", covariant=True)

class class_property:
    __fget__: Any = ...
    __doc__: Any = ...
    __name__: Any = ...
    __module__: Any = ...
    def __init__(self, getter: abc.Callable[..., _R_Co]) -> None: ...
    def __get__(self, obj: _T, typ: type = ...) -> _R_Co: ...

def try_import(modulename: str, qualname: str = ..., default: Any = ...) -> Any: ...



import typing as t

from importlib import import_module
from typing_extensions import Self


class NotSetType:
    __slots__ = ()
    __self: t.Final[Self] = None

    def __new__(cls: type[Self]) -> Self:
        self = cls.__self
        if self is None:
            self = cls.__self = super().__new__(cls)
        return self

    def __bool__(self):
        return False

    def __copy__(self, *memo):
        return self

    __deepcopy__ = __copy__

    def __reduce__(self):
        return f"{__name__}.NotSet"

    def __json__(self):
        return

    def __repr__(self):
        return "NotSet"


if t.TYPE_CHECKING:

    class NotSet(NotSetType):
        ...


NotSet = NotSetType()



import enum
import typing as t
from functools import reduce
from operator import or_
from types import DynamicClassAttribute

from typing_extensions import Self

__all__ = ["Enum", "IntEnum", "StrEnum", "Flag", "IntFlag"]

try:
    from django.db.models.enums import Choices as Enum
    from django.db.models.enums import ChoicesMeta as EnumMeta
    from django.db.models.enums import IntegerChoices as IntEnum
    from django.db.models.enums import TextChoices as StrEnum
except ImportError:

    class EnumMeta(enum.EnumMeta):
        """A metaclass for creating a enum choices."""

        def __new__(metacls, classname, bases, classdict, **kwds):
            labels = []
            for key in classdict._member_names:
                value = classdict[key]
                if (
                    isinstance(value, (list, tuple))
                    and len(value) > 1
                    and isinstance(value[-1], str)
                ):
                    *value, label = value
                    value = tuple(value)
                else:
                    label = key.replace("_", " ").title()
                labels.append(label)
                # Use dict.__setitem__() to suppress defenses against double
                # assignment in enum's classdict.
                dict.__setitem__(classdict, key, value)
            cls = super().__new__(metacls, classname, bases, classdict, **kwds)
            for member, label in zip(cls.__members__.values(), labels):
                member._label_ = label
            return enum.unique(cls)

        def __contains__(cls, member):
            if not isinstance(member, enum.Enum):
                # Allow non-enums to match against member values.
                return any(x.value == member for x in cls)
            return super().__contains__(member)

        @property
        def names(cls):
            empty = ["__empty__"] if hasattr(cls, "__empty__") else []
            return empty + [member.name for member in cls]

        @property
        def choices(cls):
            empty = [(None, cls.__empty__)] if hasattr(cls, "__empty__") else []
            return empty + [(member.value, member.label) for member in cls]

        @property
        def labels(cls):
            return [label for _, label in cls.choices]

        @property
        def values(cls):
            return [value for value, _ in cls.choices]

    class Enum(enum.Enum, metaclass=EnumMeta):
        """Class for creating enumerated choices."""

        @DynamicClassAttribute
        def label(self):
            return self._label_

        @property
        def do_not_call_in_templates(self):
            return True

        def __str__(self):
            """
            Use value when cast to str, so that Choices set as model instance
            attributes are rendered as expected in templates and similar contexts.
            """
            return str(self.value)

        # A similar format was proposed for Python 3.10.
        def __repr__(self):
            return f"{self.__class__.__qualname__}.{self._name_}"


    class IntEnum(int, Enum):
        """Class for creating enumerated integer choices."""
        pass


    class StrEnum(str, Enum):
        """Class for creating enumerated string choices."""

        def _generate_next_value_(name, start, count, last_values):
            return name



class FlagEnumMeta(EnumMeta):
   
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._member_map_:
            self._all_ = None
    
    @property
    def all(self) -> Self:
        if self._all_ is None:
            if union := reduce(or_, map(int, self._member_map_.values()), 0):
                self._all_ = self(union)
        return self._all_



class Flag(Enum, enum.Flag, metaclass=FlagEnumMeta):
    
    all: Self
    _all_: Self

    def __iter__(self):
        return (m for m in self.__class__ if m in self)

    # def __contains__(self: Self, other: Self) -> bool:
    #     return not not (self & other)

    def __len__(self) -> int:
        return f'{self:b}'.count('1')


class IntFlagChoices(Flag, enum.IntFlag):
    """Class for creating enumerated integer flag choices."""



"""A set of python utilities

"""

__version__ = 0, 0, 0

"""Proxy/PromiseProxy implementation.

This module contains critical utilities that needs to be loaded as
soon as possible, and that shall not load any third party modules.

Parts of this module is Copyright by Werkzeug Team.
"""

import sys
from abc import abstractmethod
from collections import abc
from functools import cache, reduce
from types import GenericAlias, ModuleType
import typing as t

from typing_extensions import Self, ParamSpec, Concatenate

__all__ = ("Proxy", "PromiseProxy", "try_import", "unproxy")

__module__ = __name__  # used by Proxy class body

_object_new = object.__new__
_object_setattr = object.__setattr__
_object_getattribute = object.__getattribute__
_object_delattr = object.__delattr__

_empty = object()

_P = ParamSpec("_P")
_R = t.TypeVar("_R")
_T_Co = t.TypeVar("_T_Co", covariant=True)
_T_Factory = abc.Callable[_P, _R]


def _default_cls_attr(name, type_, cls_value=_empty):
    # Proxy uses properties to forward the standard
    # class attributes __module__, __name__ and __doc__ to the real
    # object, but these needs to be a string when accessed from
    # the Proxy class directly.  This is a hack to make that work.
    # -- See Issue #1087.

    def __new__(cls, getter):

        instance = type_.__new__(cls, cls_value)
        instance.__getter = getter
        return instance

    def __get__(self, obj, cls=None):
        return self.__getter(obj) if obj is not None else self

    return type(
        name,
        (type_,),
        {
            "__new__": __new__,
            "__get__": __get__,
        },
    )


def unproxy(obj: t.Union[_R, "GetsCurrent[_R]"]) -> _R:
    """Attempt to ."""
    try:
        fn = obj._get_current_object
    except AttributeError:
        return obj
    else:
        return fn()


class GetsCurrent(t.Protocol[_R]):
    @abstractmethod
    def _get_current_object(self, *a, **kw) -> _R:
        pass

    @abstractmethod
    def __evaluated__(self) -> bool:
        pass

    @abstractmethod
    def __maybe_evaluate__(self) -> _R:
        pass

    @abstractmethod
    def __evaluate__(self) -> None:
        pass

    @abstractmethod
    def __reset_promise__(self) -> None:
        pass


class Proxy(t.Generic[_R]):
    """Proxy to another object."""

    # Code stolen from werkzeug.local.Proxy.
    __slots__ = "__local", "__args", "__kwargs", "__orig_class", "__weakref__", "__dict__"

    __local: abc.Callable[_P, _R]
    __args: tuple
    __kwargs: dict[str, t.Any]
    __bypass_names: t.Final = frozenset(["__orig_class__"])

    def __new__(
        cls,
        local: abc.Callable[_P, _R] = _empty,
        /,
        *args: _P.args,
        __name__=None,
        __doc__=None,
        **kwargs: _P.kwargs,
    ):
        self = _object_new(cls)
        if not local is _empty:
            _object_setattr(self, "_Proxy__local", local)
            _object_setattr(self, "_Proxy__args", args)
            _object_setattr(self, "_Proxy__kwargs", kwargs)

        if __name__ is not None:
            _object_setattr(self, "__custom_name__", __name__)
        if __doc__ is not None:
            _object_setattr(self, "__doc__", __doc__)

        return t.cast(t.Union[_R, GetsCurrent[_R]], self)

    @_default_cls_attr("name", str, "Proxy")
    def __name__(self):
        try:
            return self.__custom_name__
        except AttributeError:
            return self._get_current_object().__name__

    @_default_cls_attr("qualname", str, __name__)
    def __qualname__(self):
        try:
            return self.__custom_name__
        except AttributeError:
            return self._get_current_object().__qualname__

    @_default_cls_attr("module", str, __module__)
    def __module__(self):
        return self._get_current_object().__module__

    @_default_cls_attr("doc", str, __doc__)
    def __doc__(self):
        return self._get_current_object().__doc__

    # def _get_class(self):
    #     return self._get_current_object().__class__

    @property
    def __class__(self):
        return self._get_current_object().__class__

    @property
    def __orig_class__(self):
        return self._get_current_object().__orig_class__

    @__orig_class__.setter
    def __orig_class__(self, value):
        _object_setattr(self, "_Proxy__orig_class", value)

    def _get_current_object(self, *a, **kw) -> _R:
        """Get current object.

        This is useful if you want the real
        object behind the proxy at a time for performance reasons or because
        you want to pass the object into a different context.
        """
        loc: abc.Callable[_P, _R] = _object_getattribute(self, "_Proxy__local")
        if not hasattr(loc, "__release_local__"):
            return loc(*self.__args, *a, **self.__kwargs | kw)
        try:  # pragma: no cover
            # not sure what this is about
            return getattr(loc, self.__name__)
        except AttributeError:  # pragma: no cover
            raise RuntimeError(f"no object bound to {self.__name__}")

    @property
    def __dict__(self):
        try:
            return self._get_current_object().__dict__
        except RuntimeError:  # pragma: no cover
            raise AttributeError("__dict__")

    if True:

        def __repr__(self):
            try:
                obj = self._get_current_object()
            except RuntimeError:  # pragma: no cover
                return f"<{self.__class__.__name__} unbound>"
            return repr(obj)

        def __bool__(self):
            try:
                return bool(self._get_current_object())
            except RuntimeError:  # pragma: no cover
                return False

        __nonzero__ = __bool__  # Py2

        def __dir__(self):
            try:
                return dir(self._get_current_object())
            except RuntimeError:  # pragma: no cover
                return []

        def __getattr__(self, name):
            if name == "__members__":
                return dir(self._get_current_object())
            return getattr(self._get_current_object(), name)

        def __setitem__(self, key, value):
            self._get_current_object()[key] = value

        def __delitem__(self, key):
            del self._get_current_object()[key]

        def __setattr__(self, name, value):
            if name in type(self).__bypass_names:
                _object_setattr(self, name, value)
            else:
                setattr(self._get_current_object(), name, value)

        def __delattr__(self, name):
            delattr(self._get_current_object(), name)

        def __str__(self):
            return str(self._get_current_object())

        def __lt__(self, other):
            return self._get_current_object() < other

        def __le__(self, other):
            return self._get_current_object() <= other

        def __eq__(self, other):
            return self._get_current_object() == other

        def __ne__(self, other):
            return self._get_current_object() != other

        def __gt__(self, other):
            return self._get_current_object() > other

        def __ge__(self, other):
            return self._get_current_object() >= other

        def __hash__(self):
            return hash(self._get_current_object())

        def __call__(self, *a, **kw):
            return self._get_current_object()(*a, **kw)

        def __len__(self):
            return len(self._get_current_object())

        def __getitem__(self, i):
            return self._get_current_object()[i]

        def __iter__(self):
            return iter(self._get_current_object())

        def __contains__(self, i):
            return i in self._get_current_object()

        def __add__(self, other):
            return self._get_current_object() + other

        def __sub__(self, other):
            return self._get_current_object() - other

        def __mul__(self, other):
            return self._get_current_object() * other

        def __floordiv__(self, other):
            return self._get_current_object() // other

        def __mod__(self, other):
            return self._get_current_object() % other

        def __divmod__(self, other):
            return self._get_current_object().__divmod__(other)

        def __pow__(self, other):
            return self._get_current_object() ** other

        def __lshift__(self, other):
            return self._get_current_object() << other

        def __rshift__(self, other):
            return self._get_current_object() >> other

        def __and__(self, other):
            return self._get_current_object() & other

        def __xor__(self, other):
            return self._get_current_object() ^ other

        def __or__(self, other):
            return self._get_current_object() | other

        def __div__(self, other):
            return self._get_current_object().__div__(other)

        def __truediv__(self, other):
            return self._get_current_object().__truediv__(other)

        def __neg__(self):
            return -(self._get_current_object())

        def __pos__(self):
            return +(self._get_current_object())

        def __abs__(self):
            return abs(self._get_current_object())

        def __invert__(self):
            return ~(self._get_current_object())

        def __complex__(self):
            return complex(self._get_current_object())

        def __int__(self):
            return int(self._get_current_object())

        def __float__(self):
            return float(self._get_current_object())

        def __oct__(self):
            return oct(self._get_current_object())

        def __hex__(self):
            return hex(self._get_current_object())

        def __index__(self):
            return self._get_current_object().__index__()

        def __coerce__(self, other):
            return self._get_current_object().__coerce__(other)

        def __enter__(self):
            return self._get_current_object().__enter__()

        def __exit__(self, *a, **kw):
            return self._get_current_object().__exit__(*a, **kw)

        def __reduce__(self):
            return self._get_current_object().__reduce__()

        async def __aenter__(self):
            return await self._get_current_object().__aenter__()

        async def __aexit__(self, *a, **kw):
            return await self._get_current_object().__aexit__(*a, **kw)


class PartialProxy(Proxy[_R]):
    __slots__ = ()

    def __call__(self, *a, **kw):
        return self._get_current_object(*a, **kw)


class PromiseProxy(Proxy[_R]):
    """Proxy that evaluates object once.

    :class:`Proxy` will evaluate the object each time, while the
    promise will only evaluate it once.
    """

    __slots__ = "__pending", "__thing"

    def _get_current_object(self) -> _R:
        try:
            return _object_getattribute(self, "_PromiseProxy__thing")
        except AttributeError:
            return self.__evaluate__()

    def __then__(self, fun, *args, **kwargs):
        if self.__evaluated__():
            return fun(*args, **kwargs)
        from collections import deque

        try:
            pending = _object_getattribute(self, "_PromiseProxy__pending")
        except AttributeError:
            pending = None
        if pending is None:
            pending = deque()
            _object_setattr(self, "_PromiseProxy__pending", pending)
        pending.append((fun, args, kwargs))

    def __evaluated__(self):
        try:
            _object_getattribute(self, "_PromiseProxy__thing")
        except AttributeError:
            return False
        return True

    def __maybe_evaluate__(self) -> _R:
        return self._get_current_object()

    def __reset_promise__(self) -> None:
        try:
            _object_delattr(self, "_PromiseProxy__thing")
        except AttributeError:
            pass

    def __evaluate__(self, _clean: t.Union[bool, abc.Sequence[str]] = None) -> _R:
        try:
            thing = Proxy._get_current_object(self)
        except Exception:
            raise
        else:
            _object_setattr(self, "_PromiseProxy__thing", thing)
            # if _clean is True:
            #     _clean = ("_Proxy__local", "_Proxy__args", "_Proxy__kwargs")
            # for attr in _clean or ():
            #     try:
            #         _object_delattr(self, attr)
            #     except AttributeError:  # pragma: no cover
            #         # May mask errors so ignore
            #         pass
            try:
                pending = _object_getattribute(self, "_PromiseProxy__pending")
            except AttributeError:
                pass
            else:
                # try:
                while pending:
                    fun, args, kwargs = pending.popleft()
                    fun(*args, **kwargs)
                # finally:
                #     try:
                #         _object_delattr(self, "_PromiseProxy__pending")
                #     except AttributeError:  # pragma: no cover
                #         pass
            return thing



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zana', 'zana.common', 'zana.proxy', 'zana.types']

package_data = \
{'': ['*']}

install_requires = \
['typing-extensions>=4.4.0,<5.0.0']

setup_kwargs = {
    'name': 'zana',
    'version': '0.0.2',
    'description': 'general utility functions and types',
    'long_description': '# Python Zana\n\n\n[![PyPi version][pypi-image]][pypi-link]\n[![Supported Python versions][pyversions-image]][pyversions-link]\n[![Build status][ci-image]][ci-link]\n[![Coverage status][codecov-image]][codecov-link]\n\n\nA Python tool kit\n\n\n\n## Installation\n\nInstall from [PyPi](https://pypi.org/project/zana/)\n\n```\npip install zana\n```\n\n## Documentation\n\nFull documentation is available [here][docs-link].\n\n\n\n## Production\n\n__This package is still in active development and should not be used in production environment__\n\n\n\n\n[docs-link]: https://python-zana.github.io/zana/\n[pypi-image]: https://img.shields.io/pypi/v/zana.svg?color=%233d85c6\n[pypi-link]: https://pypi.python.org/pypi/zana\n[pyversions-image]: https://img.shields.io/pypi/pyversions/zana.svg\n[pyversions-link]: https://pypi.python.org/pypi/zana\n[ci-image]: https://github.com/python-zana/zana/actions/workflows/workflow.yaml/badge.svg?event=push&branch=main\n[ci-link]: https://github.com/python-zana/zana/actions?query=workflow%3ACI%2FCD+event%3Apush+branch%3Amaster\n[codecov-image]: https://codecov.io/gh/python-zana/zana/branch/main/graph/badge.svg\n[codecov-link]: https://codecov.io/gh/python-zana/zana\n\n\nSee this release on GitHub: [v0.0.2](https://github.com/python-zana/zana/releases/tag/0.0.2)\n',
    'author': 'David Kyalo',
    'author_email': 'davidmkyalo@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/python-zana/zana',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
